/*BEGIN CMVC */
//*************************************************************************
//
//  Workfile: %W%
//  Last update: %E% %U%
//  SCCS path, id: %F% %I%
//
//************************************************************************/
/*END CMVC */
/*BEGIN COPYRIGHT */
//*************************************************************************
//
//  Licensed Materials - Property of IBM
//  WA82 SupportPac
//  (c) Copyright IBM Corp. 2000, 2001, 2002
//  US Government Users Restricted Rights - Use, duplication or
//  disclosure restricted by GSA ADP Schedule Contract
//  with IBM Corp.
//
//************************************************************************/
/*END COPYRIGHT */
/*BEGIN DISCLAIMER */
//*************************************************************************
//  DISCLAIMER
//  ----------
//  This material contains programming source code for your consideration.
//  These examples have not been thoroughly tested under all conditions.
//  IBM, therefore, cannot guarantee or imply reliability, serviceability,
//  or function of these programs.
//  ALL PROGRAMS CONTAINED HEREIN ARE PROVIDED TO YOU "AS IS", WITHOUT ANY
//  WARRANTIES (EXPRESS OR IMPLIED) OR SUPPORT WHATSOEVER, INCLUDING BUT
//  NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
//  FOR A PARTICULAR PURPOSE.
//************************************************************************/
/*END DISCLAIMER */

package com.ibm.workflow.sample.upes;

import java.io.*;
import java.util.*;

import com.ibm.mq.*;

import org.apache.xerces.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

public abstract class UPESBaseSAX extends DefaultHandler
{
  protected final static String eol = "\r\n";
  private   final static int NO_CONTAINER  = 1;
  private   final static int IN_CONTAINER  = 2;
  private   final static int OUT_CONTAINER = 3;
  private   final static int AUDIT_EVENT   = 4;

  private MQQueue        inQueue;
  private MQQueue        outQueue;
  private MQQueueManager inQueueManager;
  private MQQueueManager outQueueManager;
  private SAXParser      saxParser;

  // Maps to hold the contents of the message
  private Map header;
  private Map inContainer;
  private Map outContainer;
  private Map auditEvent;

  // Temporary (global) variables for SAX2 parsing
  private String       latestTag;
  private StringBuffer currentElement;
  private int          currentContainer;

  // UPESBaseSAX Constructor
  // This creates a UPES that listens on the specified input queue on the
  // specified input queue manager
  // ----------------------------------------------------------------------
  public UPESBaseSAX( String inQueueManagerName, String inQueueName,
                      String mqHostname, String mqPort, String mqChannel )
  {
    System.out.println( "UPESBaseSAX initializing" );

    try {
      System.out.println( "  Connecting to Queue Manager '" + inQueueManagerName +"'" );
      if ( mqHostname != null )
          MQEnvironment.hostname = mqHostname;
      if ( mqPort != null )
          MQEnvironment.port = Integer.parseInt( mqPort );
      if ( mqChannel != null )
          MQEnvironment.channel = mqChannel;
      
      inQueueManager = new MQQueueManager( inQueueManagerName );

      System.out.println( "  Accessing queue " + inQueueName );
      int openOptions = MQC.MQOO_INPUT_SHARED;
      inQueue = inQueueManager.accessQueue( inQueueName, openOptions );

      saxParser = new SAXParser( );
      saxParser.setContentHandler( this );

      System.out.println( "UPESBaseSAX initialized" );
      System.out.println( );
    }
    catch( Exception xcpt ) {
      xcpt.printStackTrace( );
      System.exit( 1 );
    }
  }

  // Shutdown the Base UPES. This function should be called from the
  // derived class upon application shutdown.
  // It closes all connections and disconnects from all queue managers.
  // ----------------------------------------------------------------------
  protected void shutdown( )
  {
    System.out.println( "UPESBaseSAX shutdown" );

    // Disconnect from MQSeries
    try {
      if ( inQueue  != null ) inQueue.close( );
      if ( outQueue != null ) outQueue.close( );
      if ( inQueueManager  != null ) inQueueManager.disconnect( );
      if ( ( outQueueManager != null ) && (outQueueManager != inQueueManager ) )
        outQueueManager.disconnect( );
    }
    catch( Exception xcpt ) {
      xcpt.printStackTrace( );
    }
    finally {
      inQueue         = null;
      outQueue        = null;
      inQueueManager  = null;
      outQueueManager = null;
    }

    System.out.println( "UPESBaseSAX shutdown complete" );
    System.out.println( );
  }

  // Main loop of the UPES. This loop waits for the next message,
  // checks for a QUIT message, decodes the message and finally sends
  // the response message back to MQWF (if required)
  //
  // This method has to be called by the derived class to start processing
  // ----------------------------------------------------------------------
  public void run( )
  {
    do {
      try {
        MQMessage inMessage = new MQMessage( );
        MQGetMessageOptions gmo = new MQGetMessageOptions( );
        gmo.options      = MQC.MQGMO_WAIT | MQC.MQGMO_SYNCPOINT;
        gmo.waitInterval = MQC.MQWI_UNLIMITED;

        // Wait for the next message and get it
        System.out.println( "-----------------------------------------------------------" );
        System.out.println( "Waiting for next message" );
        inQueue.get( inMessage, gmo );
        System.out.println( "Message received" );

        if ( inMessage.getTotalMessageLength( ) == 4 ) {
          // QUIT Message received
          inQueueManager.commit( );
          break;
        }
        else {
          header       = Collections.synchronizedMap( new HashMap( ) );
          inContainer  = Collections.synchronizedMap( new HashMap( ) );
          outContainer = Collections.synchronizedMap( new HashMap( ) );
          auditEvent   = Collections.synchronizedMap( new HashMap( ) );

          // Call our XML Parsing method
          decodeMessage( inMessage );
        }

        // Check if Audit Trail Record
        if ( auditEvent.size( ) > 0 ) {
          processAuditEventMessage( );
        }
        else {
          // Check if MQWF expects a response message
          if ( ((String)header.get( "ResponseRequired" )).equalsIgnoreCase( "yes" ) ) {
            // Process the message (in the derived class)
            System.out.println( "Response required" );
            String xmlOutputContainer = processSyncActivityInvokeMessage( );

            sendResponseMessage( inMessage, xmlOutputContainer, returnCode( ) );
            if ( ( outQueueManager != null ) && ( outQueueManager != inQueueManager ) )
              outQueueManager.commit( );
          }
          else processAsyncActivityInvokeMessage( );
        }
        
        inQueueManager.commit( );
      }
      catch( Exception xcpt ) {
        xcpt.printStackTrace( );
        try {
          inQueueManager.backout( );
        } catch( MQException e ) { }
      }
    } while ( true );

    System.out.println( "QUIT received" );
    System.out.println( );
  }

  // Send the response message back to MQSeries Workflow. This method will
  // only be invoked if ResponseRequired is set to YES.
  // ----------------------------------------------------------------------
  private void sendResponseMessage( MQMessage inMessage, String xmlOutputContainer, int RC )
    throws MQException, IOException
  {
    if ( outQueueManager == null ) {
      System.out.print( "Accessing outgoing QueueManager: " +
                        inMessage.replyToQueueManagerName );
      // Check to see if the outgoing queue manager is the same as
      // the incoming queue manager. If yes -> reuse it. Better transaction
      // coordination as well (one commit instead of two)
      if ( inQueueManager.name.trim( ).equals( inMessage.replyToQueueManagerName.trim( ) ) ) {
        System.out.println( " (Reusing incoming Queue Manager)" );
        outQueueManager = inQueueManager;
      }
      else {
        System.out.println( " (Connecting to new Queue Manager)" );
        outQueueManager = new MQQueueManager( inMessage.replyToQueueManagerName );
      }
    }
    if ( outQueue == null ) {
      int openOptions = MQC.MQOO_OUTPUT;
      System.out.println( "Accessing outgoing Queue: " +
                          inMessage.replyToQueueName.trim( ) );
      outQueue = outQueueManager.accessQueue( inMessage.replyToQueueName.trim( ),
                                              openOptions, null, null, null );
    }

    MQMessage outMessage = new MQMessage( );
    outMessage.messageType   = MQC.MQMT_REPLY;
    outMessage.messageId     = inMessage.messageId;
    outMessage.correlationId = inMessage.correlationId;

    StringBuffer msgText = new StringBuffer( 1024 );
    msgText.append( "<?xml version=\"1.0\" standalone=\"yes\"?>" + eol );
    msgText.append( "<!-- This document generated by UPESBaseSAX class. -->" + eol );
    msgText.append( "<!-- (w) 2000 Wolfgang Kulhanek. -->" + eol );
    msgText.append( "<WfMessage>" + eol );
    msgText.append( "<ActivityImplInvokeResponse>" + eol );
    msgText.append( "<ActImplCorrelID>" + (String)header.get( "ActImplCorrelID" ) +
                    "</ActImplCorrelID>" + eol );
    msgText.append( "<ProgramRC>" + RC + "</ProgramRC>" + eol );

    // Now add the container data
    msgText.append( "<ProgramOutputData>" + eol );
    msgText.append( xmlOutputContainer );
    msgText.append( "</ProgramOutputData>" + eol );
    msgText.append( "</ActivityImplInvokeResponse>" + eol );
    msgText.append( "</WfMessage>" );

    outMessage.writeString( msgText.toString( ) );

    MQPutMessageOptions pmo = new MQPutMessageOptions( );
    pmo.options = MQC.MQPMO_SYNCPOINT;

    outQueue.put( outMessage, pmo );
  }

  // Convenience Function to dump a container (Map) to the screen
  // ----------------------------------------------------------------------
  public void dumpContainer( Map container )
  {
    for ( Iterator i = container.entrySet( ).iterator( ); i.hasNext( ); ) {
      Map.Entry entry = (Map.Entry) i.next( );
      System.out.println( entry.getKey( ) + ": " + entry.getValue( ) );
    }
  }

  // Convenience Function to read a String (default: "") from a container
  // -----------------------------------------------------------------------
  public final String getString( Map container, String key )
  {
    String value = (String)container.get( key );
    if ( value == null ) return "";
    else                 return value;
  }

  // Convenience Function to read a long (default: 0) from a container
  // ----------------------------------------------------------------------
  public final long getLong( Map container, String key )
  {
    long value = 0;
    try {
      value = new Long( (String)container.get( key )).longValue( );
    } catch( Exception xcpt ) { /* Ignore exception - return 0 */ }
    return value;
  }

  // Convenience Function to read a double (default: 0.0) from a container
  // ----------------------------------------------------------------------
  public final double getDouble( Map container, String key )
  {
    double value = 0.0;
    try {
      // Replace ',' with '.' in case ',' is used as a decimal separator
      // If this is not done, doubleValue( ) will throw an exception if
      // the decimal separator is set to ','.
      value = new Double( ((String)container.get( key )).replace( ',', '.' )).doubleValue( );
    } catch( Exception xcpt ) { /* Ignore exception - return 0.0 */ }
    return value;
  }

  // Public access to the various containers
  // Header (everything outside the in- and outcontainer)
  // Input Container
  // Output Container Defaults
  // ----------------------------------------------------------------------
  public final Map getHeader( ) {       return header;       }
  public final Map getInContainer( ) {  return inContainer;  }
  public final Map getOutContainer( ) { return outContainer; }
  public final Map getAuditEvent( ) {   return auditEvent;   }

  
  // XML/SAX2 Parsing methods
  // ----------------------------------------------------------------------

  private void decodeMessage( MQMessage msg )
    throws IOException, SAXException
  {
    byte[] b = new byte[ msg.getDataLength( ) ];
    msg.readFully( b, 0, msg.getDataLength( ) );

    ByteArrayInputStream bais = new ByteArrayInputStream( b );
    InputSource is = new InputSource( bais );

    try {
      currentContainer = NO_CONTAINER;
      currentElement   = new StringBuffer( 1024 );
      saxParser.parse( is );

      // Make collection immutable
      header       = Collections.unmodifiableMap( header       );
      inContainer  = Collections.unmodifiableMap( inContainer  );
      outContainer = Collections.unmodifiableMap( outContainer );
      auditEvent   = Collections.unmodifiableMap( auditEvent   );
    }
    catch( SAXException xcpt ) {
      xcpt.printStackTrace( );
      header       = null;
      inContainer  = null;
      outContainer = null;
      auditEvent   = null;
    }
    bais.close( );
  }

  // This functions override the methods in ContentHandler for SAX2 parsing
  // ----------------------------------------------------------------------

  public void startElement( String namespaceURI, String localName, String qName, Attributes attrs )
  {
    // Store the name of the tag & create a new StringBuffer
    latestTag = qName;

    if ( latestTag.equals( "AuditTrailRecord" ) )
      currentContainer = AUDIT_EVENT;
    if ( ( currentContainer==NO_CONTAINER ) && ( latestTag.equals( "ProgramInputData" ) ) )
      currentContainer = IN_CONTAINER;
    if ( ( currentContainer==IN_CONTAINER ) && ( latestTag.equals( "ProgramOutputDataDefaults" ) ) )
      currentContainer = OUT_CONTAINER;
  }

  public void characters( char[] ch, int start, int length )
  {
    currentElement.append( new String( ch, start, length ) );
  }

  public void endElement( String namespaceURI, String localName, String qName )
  {
    String currentString = currentElement.toString( ).trim( );
    if ( currentString.length( ) > 0 ) {
      switch( currentContainer ) {
        case NO_CONTAINER:
          header.put( latestTag, currentString );
          break;
        case IN_CONTAINER:
          inContainer.put( latestTag, currentString );
          break;
        case OUT_CONTAINER:
          outContainer.put( latestTag, currentString );
          break;
        case AUDIT_EVENT:
          auditEvent.put( latestTag, currentString );
          break;
      }
      currentElement.setLength( 0 );
    }
  }

  // The following methods are declared abstract and have to be implemented
  // by the class that is derived from UPESBaseSAX

  // Callback method for synchronous UPES activities.
  // This method should return the output data container as an XML
  // String without the enclosing <ProgramOutputData> and </ProgramOutputData>
  // tags.
  // ----------------------------------------------------------------------
  public abstract String processSyncActivityInvokeMessage( );

  // Callback method for asynchronous UPES activities.
  // This method doesn't need to return any XML data since there is no output
  // container.
  // ----------------------------------------------------------------------
  public abstract void processAsyncActivityInvokeMessage( );

  // Callback method for expired UPES activities.
  // This method doesn't need to return any XML data since there is no output
  // container.
  // ----------------------------------------------------------------------
  public abstract void processActivityTerminatedMessage( );

  // Callback method for terminated UPES activities.
  // This method doesn't need to return any XML data since there is no output
  // container.
  // ----------------------------------------------------------------------
  public abstract void processActivityExpiredMessage( );

  // Callback method for audit event messages.
  // This method doesn't need to return any XML data since there is no output
  // container.
  // ----------------------------------------------------------------------
  public abstract void processAuditEventMessage( );

  // Callback method that should return the returncode for the activity.
  // ----------------------------------------------------------------------
  public abstract int returnCode( );
}
